{
module Parse  where

-- import Alex

import Lex
import Tokens
import AbsSyn
import Debug.Trace(trace)
}

%tokentype { (Token AlexPosn) }

%name parse

%token
  	'{'             { LBrace _ }
  	'}'		{ RBrace _ }
  	'['		{ LBrack _ }
  	']'		{ RBrack _ }
  	'('		{ LPar _ }
  	')'		{ RPar _ }
   	'<-' 		{ Ass _ } 
  	'?'             { Ife _ }
  	':' 		{ Colon _ }
  	',' 		{ Comma _ }
  	';' 		{ Semicol _ }
  	'&&'		{ And _ } 
  	'^^'		{ Xor _ } 
  	'||'		{ Or _ } 
  	'<<'		{ Lshift _ } 
  	'<<#'		{ Ashift _ } 
  	'>>'		{ Rshift _ } 
   	'=' 		{ Def _ } 
   	'==' 		{ Eq _ } 
   	'!=' 		{ Ne _ } 
   	'<'		{ Lt _ } 
   	'<='		{ Le _ } 
   	'>'		{ Gt _ } 
   	'>='		{ Ge _ } 
   	'<#'		{ ULt _ } 
   	'<=#'		{ ULe _ } 
   	'>#'		{ UGt _ } 
   	'>=#'		{ UGe _ } 
  	'+' 		{ Plus _ }
  	'-'		{ Minus _ } 
  	'*' 		{ Times _ }
  	'/' 		{ Div _ }
  	SET             { Set _ }
  	NAME            { Name _ }
  	VALUE           { Value _ }
  	EXP             { Exp _ }
  	MEM             { Mem _ }
  	SEQ             { Seq _ }
  	CALL            { Call _ }
  	FS              { FS _ }
  	FCT             { Tokens.Fct _ }
  	REG             { Reg _ }
  	LOC             { Loc _ }
  	ADR             { Adr _ }
  	JUMP            { Jump _ }
  	CJUMP           { CJump _ }
        integer         { Const $$ _ }
        label           { Label $$ _ }
        temp            { Register $$ _ }
        identifier      { Id $$ _ }


%left '<-'
%left '||' '^^'
%left '&&'
%left '<' '<=' '>' '>=' '==' '!='
%left '<#' '<=#' '>#' '>=#'
%left '<<' '<<#' '>>'
%left '+'  '-'
%left '*'  '/'
%left NEG

%%

Prg :: { ([(String, Either String Integer)] , [AbsSyn.Fct]) }
Prg: DefList FctList           { ($1, $2) }

FctList :: { [AbsSyn.Fct] }
FctList: {- empty -} {[]}
  | Fct FctList { $1:$2 }

Fct :: { AbsSyn.Fct }
Fct: FCT label '(' ParamList ')' MaybeFrameSize '=' '{' StmList '}' { AbsSyn.Fct $2 $4 $6 $9 }

DefList :: { [(String, Either String Integer)] }
DefList: {- empty -} {[]}
  | Def DefList { $1:$2 }

Def :: { (String, Either String Integer) }
Def:  SET NAME identifier '=' identifier { ($3,Left $5) }
  |   SET VALUE identifier '=' integer   { ($3,Right $5) }
  |   SET NAME identifier '=' temp       { ($3,Left $5) }

StmList :: { [AbsSyn.Stm] }
StmList: {- empty -}    { [] }
  | Stm StmList0       { $1:$2 }

StmList0 :: { [AbsSyn.Stm] }
StmList0: {- empty -}   { [] }
  | Stm StmList0   { $1:$2 }

Stm :: { AbsSyn.Stm }
Stm:  Exp '<-' Exp                { AbsSyn.MOVE $1 $3 }
|  Exp                        { AbsSyn.EXP $1 }
|  JUMP Exp '[' LabelList ']'     { AbsSyn.JUMP $2 $4 }
|  CJUMP '(' RelExpr ')' '?'  label ':'  label { AbsSyn.CJUMP (fst3 $3) (snd3 $3) (thd3 $3) $6 $8 }
|  SEQ '(' Stm ',' Stm ')'     { AbsSyn.SEQ $3 $5 }
|  label ':'                   { AbsSyn.LABEL $1 }

Integer :: { Integer }
Integer : integer                { $1}
        | '-' integer %prec NEG  { - $2 }

Exp :: { AbsSyn.Exp }
Exp: label          { AbsSyn.NAME $1 }
| temp              { AbsSyn.TEMP $1 }
| Integer           { AbsSyn.CONST $1 }
| InfxExp    { $1 }
| '(' Exp ')'       { $2 }
| MEM  '(' Exp ')'           { AbsSyn.MEM $3 }
| CALL Exp '(' ExpList ')'      { AbsSyn.CALL $2 $4 }
| '{' StmList ',' Exp '}' { AbsSyn.ESEQ (mkSeq $2) $4 }

ExpList :: { [AbsSyn.Exp] }
ExpList: {- empty -}    { [] }
  |  Exp ExpList0        { $1:$2 }

ExpList0 :: { [AbsSyn.Exp] }
ExpList0: {- empty -}    { [] }
  |  ',' Exp ExpList0        { $2:$3 }

LabelList :: { [AbsSyn.Label] }
LabelList: {- empty -}    { [] }
  | label LabelList0       { $1:$2 }

LabelList0 :: { [AbsSyn.Label] }
LabelList0: {- empty -}   { [] }
  | ',' label LabelList0   { $2:$3 }

Param: LOC integer   { AbsSyn.LOC $2 }
  | REG temp         { AbsSyn.REG $2 }
  | ADR integer      { AbsSyn.ADR $2 }

ParamList :: { [AbsSyn.Param] }
ParamList: {- empty -}    { [] }
  | Param ParamList0       { $1:$2 }

ParamList0 :: { [AbsSyn.Param] }
ParamList0: {- empty -}   { [] }
  | ',' Param ParamList0   { $2:$3 }

MaybeFrameSize :: { Maybe Integer }
MaybeFrameSize: {- empty -}   { Nothing }
  | FS integer                { Just $2 }


InfxExp :: { AbsSyn.Exp }
InfxExp: Exp '*' Exp     { AbsSyn.BINOP AbsSyn.MUL $1 $3 }
| Exp '+' Exp     { AbsSyn.BINOP AbsSyn.PLUS $1 $3 }
| Exp '-' Exp     { AbsSyn.BINOP AbsSyn.MINUS $1 $3 }
| Exp '/' Exp     { AbsSyn.BINOP AbsSyn.DIV $1 $3 }
| Exp '&&' Exp     { AbsSyn.BINOP AbsSyn.AND $1 $3 }
| Exp '||' Exp     { AbsSyn.BINOP AbsSyn.OR $1 $3 }
| Exp '^^' Exp     { AbsSyn.BINOP AbsSyn.XOR $1 $3 }
| Exp '<<' Exp     { AbsSyn.BINOP AbsSyn.LSHIFT $1 $3 }
| Exp '>>' Exp     { AbsSyn.BINOP AbsSyn.RSHIFT $1 $3 }
| Exp '<<#' Exp     { AbsSyn.BINOP AbsSyn.ARSHIFT $1 $3 }

RelExpr :: { (AbsSyn.RelOp, AbsSyn.Exp, AbsSyn.Exp) }
RelExpr: Exp '==' Exp   { (AbsSyn.EQ, $1, $3) }
     | Exp '!=' Exp   {  (AbsSyn.NE, $1, $3) }
     | Exp '<' Exp   {  (AbsSyn.LT, $1, $3) }
     | Exp '>' Exp   {  (AbsSyn.GT, $1, $3) }
     | Exp '<=' Exp   {  (AbsSyn.LE, $1, $3) }
     | Exp '>=' Exp   {  (AbsSyn.GE, $1, $3) }
     | Exp '<#' Exp   {  (AbsSyn.ULT, $1, $3) }
     | Exp '>#' Exp   {  (AbsSyn.UGT, $1, $3) }
     | Exp '<=#' Exp   {  (AbsSyn.ULE, $1, $3) }
     | Exp '>=#' Exp   {  (AbsSyn.UGE, $1, $3) }


{
fst3 (x,_,_) = x
snd3 (_,y,_) = y
thd3 (_,_,z) = z

mkSeq :: [Stm] -> Stm
mkSeq [] = error "mkSeq: empty list"
mkSeq [x] = x
mkSeq (x:xs) = AbsSyn.SEQ x (mkSeq xs)

-- toOperation :: (Show a) => Token a -> Operation */
-- toOperation (OpC x _) = toEnum (fromEnum x) */
-- toOperation x = error $ "illegal token " ++ (show x) ++ " found instead of an OpCode" */

--happyError :: [Token AlexPosn] -> a
--happyError pos = error $ "Parse error at" ++ (show pos)

happyError :: [Token AlexPosn] -> a
happyError tks = error ("Parse error at " ++ lcn ++ "\n")
	where
	lcn = 	case tks of
		  [] -> "end of file"
		  (tk:_) -> "line " ++ show l ++ ", column " ++ show c ++ " (token " ++ filterPn (show tk) ++ ")"
			where AlexPn _ l c = token_pos tk
  {-
| Exp '+' Exp { AbsSyn.BINOP AbsSyn.PLUS $1 $3 }
| Exp '-' Exp { AbsSyn.BINOP AbsSyn.MINUS $1 $3 }
| Exp '*' Exp { AbsSyn.BINOP AbsSyn.MUL $1 $3 }

BinOpExp :: { AbsSyn.Exp }
BinOpExp: Exp BinOp Exp     { AbsSyn.BINOP $2 $1 $3 }


BinOp :: { AbsSyn.BinOp }
BinOp: '+'   { AbsSyn.PLUS }
     | '-'   { AbsSyn.MINUS }
     | '*'   { AbsSyn.MUL }
     | '/'   { AbsSyn.DIV }
     | '&&'  { AbsSyn.AND }
     | '||'  { AbsSyn.OR }
     | '^^'  { AbsSyn.XOR }


   -}
}
